"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.drawTextWithEmoji = exports.textWidth = exports.getBaseline = exports.getEmojiSize = exports.getLineHeight = exports.getAbsoluteFontSize = exports.getLines = exports.fixLine = void 0;
const canvas_1 = require("canvas");
const emoji_1 = require("./emoji");
const cachedEmojiImages = new Map();
async function loadEmojiImageByUrl(url) {
    if (cachedEmojiImages.has(url)) {
        return cachedEmojiImages.get(url);
    }
    const image = await (0, canvas_1.loadImage)(url);
    cachedEmojiImages.set(url, image);
    return image;
}
const DISCORD_EMOJI_PATTERN = /<a?:\w+:(\d{17,19})>/g;
const DISCORD_EMOJI_PATTERN_SINGLE = /<a?:\w+:(\d{17,19})>/;
const EMOJI_REPLACEMENT_CHAR = "‱";
const EMOJI_SIDE_MARGIN = 0.1;
function fixLine(str) {
    return str.trim().replace(/\s{2,}/g, " ");
}
exports.fixLine = fixLine;
function getLines(ctx, text, maxWidth) {
    if (maxWidth < 10) {
        throw new Error("getLines: invalid max width");
    }
    let lines = [];
    let emojis = text.matchAll(DISCORD_EMOJI_PATTERN);
    let textLines = text
        .replace(DISCORD_EMOJI_PATTERN, EMOJI_REPLACEMENT_CHAR)
        .split("\n");
    for (const textLine of textLines) {
        let line = "";
        let words = textLine.split(" ");
        for (let word of words) {
            word = fixLine(word);
            if (!word) {
                continue;
            }
            if (textWidth(ctx, word) > maxWidth) {
                lines.push(line);
                line = "";
                let width = 0;
                let i = 0;
                let l = 1;
                while (i + l < word.length) {
                    while (width < maxWidth && i + l < word.length) {
                        width = textWidth(ctx, word.substring(i, i + l - 1));
                        l++;
                    }
                    if (i + l >= word.length) {
                        lines.push(word.substring(i, i + l));
                    }
                    else {
                        lines.push(word.substring(i, i + l - 1));
                    }
                    i += l - 1;
                    l = 1;
                    width = 0;
                }
                continue;
            }
            const testLine = fixLine(line + " " + word);
            if (textWidth(ctx, testLine) > maxWidth) {
                lines.push(line);
                line = word;
            }
            else {
                line = testLine;
            }
        }
        lines.push(line);
    }
    if (emojis) {
        let text = lines.join("\n");
        for (const emoji of emojis) {
            text = text.replace(EMOJI_REPLACEMENT_CHAR, `\u200b${emoji[0]}\u200b`);
        }
        lines = text.split("\n");
    }
    return lines.map((line) => fixLine(line)).filter((line) => !!line);
}
exports.getLines = getLines;
function parseDiscordEmojis(textEntities) {
    const newTextEntities = [];
    for (const entity of textEntities) {
        if (typeof entity === "string") {
            for (const word of entity.split("\u200b")) {
                const match = word.match(DISCORD_EMOJI_PATTERN_SINGLE);
                newTextEntities.push(match
                    ? { url: `https://cdn.discordapp.com/emojis/${match[1]}.png` }
                    : word);
            }
        }
        else {
            newTextEntities.push(entity);
        }
    }
    return newTextEntities;
}
function splitEntitiesFromText(text) {
    const emojiEntities = (0, emoji_1.parse)(text);
    let unparsedText = text;
    let lastEmojiIndice = 0;
    const textEntities = [];
    emojiEntities.forEach((emoji) => {
        textEntities.push(unparsedText.slice(0, emoji.indices[0] - lastEmojiIndice));
        textEntities.push(emoji);
        unparsedText = unparsedText.slice(emoji.indices[1] - lastEmojiIndice);
        lastEmojiIndice = emoji.indices[1];
    });
    textEntities.push(unparsedText);
    return parseDiscordEmojis(textEntities);
}
function getAbsoluteFontSize(ctx) {
    const fontMetrics = ctx.measureText("HELLO WORLD");
    return (fontMetrics.actualBoundingBoxAscent + fontMetrics.actualBoundingBoxDescent);
}
exports.getAbsoluteFontSize = getAbsoluteFontSize;
function getLineHeight(ctx) {
    return 1.5 * getAbsoluteFontSize(ctx);
}
exports.getLineHeight = getLineHeight;
function getEmojiSize(ctx) {
    return Math.round(getLineHeight(ctx));
}
exports.getEmojiSize = getEmojiSize;
function getBaseline(ctx) {
    const fontMetrics = ctx.measureText("HELLO WORLD");
    return fontMetrics.alphabeticBaseline;
}
exports.getBaseline = getBaseline;
function textWidth(context, text) {
    const textEntities = splitEntitiesFromText(text);
    const emojiSize = getEmojiSize(context);
    const emojiSideMargin = emojiSize * EMOJI_SIDE_MARGIN;
    let currentWidth = 0;
    for (let i = 0; i < textEntities.length; i++) {
        const entity = textEntities[i];
        if (typeof entity === "string") {
            currentWidth += context.measureText(entity).width;
        }
        else {
            currentWidth += emojiSize + emojiSideMargin * 2;
        }
    }
    return currentWidth;
}
exports.textWidth = textWidth;
async function drawTextWithEmoji(context, fillType, text, x, y, drawWidth, textAlign) {
    const textEntities = splitEntitiesFromText(text);
    const emojiSize = getEmojiSize(context);
    const emojiSideMargin = emojiSize * EMOJI_SIDE_MARGIN;
    context.textBaseline = "top";
    const width = textWidth(context, text);
    //text align
    let textLeftMargin = 0;
    context.textAlign = "left";
    switch (textAlign) {
        case "center":
            textLeftMargin = drawWidth / 2 - width / 2;
            break;
        case "end":
        case "right":
            textLeftMargin = drawWidth - width;
            break;
    }
    let currentWidth = 0;
    for (let i = 0; i < textEntities.length; i++) {
        const entity = textEntities[i];
        if (typeof entity === "string") {
            if (fillType === "fill") {
                context.fillText(entity, textLeftMargin + x + currentWidth, y);
            }
            else {
                context.strokeText(entity, textLeftMargin + x + currentWidth, y);
            }
            currentWidth += context.measureText(entity).width;
        }
        else {
            try {
                const emoji = await loadEmojiImageByUrl(entity.url);
                if (!emoji) {
                    continue;
                }
                const emojiHeight = Math.round((emojiSize / emoji.width) * emoji.height);
                const offsetY = Math.max(0, Math.round((emojiSize - emojiHeight) / 2));
                context.drawImage(emoji, textLeftMargin + x + currentWidth + emojiSideMargin, y + offsetY, emojiSize, emojiHeight);
            }
            catch { }
            currentWidth += emojiSize + emojiSideMargin * 2;
        }
    }
    if (textAlign) {
        context.textAlign = textAlign;
    }
}
exports.drawTextWithEmoji = drawTextWithEmoji;
